#!/usr/bin/env python
############################################################################
#    Copyright (C) 2004 by Christian Berg                                  #
#    cberg@users.sourceforge.net                                           #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################
import sys, os, pygame, random
import pprint, profile, pstats
from pygame.locals import *


def load_image(name, colorkey=None):
    fullname = os.path.join('data', name)
    try:
        image = pygame.image.load(fullname)
    except pygame.error, message:
        print 'Cannot load image:', fullname
        raise SystemExit, message
    image = image.convert()
    if colorkey is not None:
        if colorkey is -1:
            colorkey = image.get_at((0,0))
        image.set_colorkey(colorkey,RLEACCEL)
    return image, image.get_rect()

def distance(pos1,pos2):
    (x1,y1) = pos1
    (x2,y2) = pos2
    return abs(x1-x2) + abs(y1-y2)




class PyPebble(pygame.sprite.Sprite):
    def __init__(self, type, x, y):
        global stones
        pygame.sprite.Sprite.__init__(self)
        self.image, rect = stones[type]
        self.rect = Rect(rect)
        self.type, self.x, self.y = type, x, y
        self.posx, self.posy = 50.0*x, -50.0
        self.velx, self.vely = 10.0, 10.0
        self.dirx, self.diry = 1, 1
        self.deleted = 0
    def set_field(self,x,y):
        self.x, self.y = x, y
        if x*50 < self.posx:
            self.dirx = -1
        else:
            self.dirx = 1
        if y*50 < self.posy:
            self.diry = -1
        else:
            self.diry = 1
    def update(self):
        time_elapsed=1
        if self.posx == self.x*50 and self.posy == self.y*50:
            return 0
        self.posx += self.dirx*self.velx
        self.posy += self.diry*self.vely
        if (self.posx-self.x*50)/self.dirx >= 0:
            self.posx = self.x*50
        if (self.posy-self.y*50)/self.diry >= 0:
            self.posy = self.y*50
        self.rect.topleft = (self.posx,self.posy)
        return 1

class MyRenderGroup(pygame.sprite.RenderUpdates):
    def update(self, *args):
        ret_combined = 0
        if args:
            a=apply
            for s in self.spritedict.keys():
                retval = a(s.update, args)
                ret_combined = ret_combined or retval
        else:
            for s in self.spritedict.keys():
                retval = s.update()
                ret_combined = ret_combined or retval
        return ret_combined

class PyPebblesGame:
    def __init__(self):
        pygame.init()
        self.screen = pygame.display.set_mode((400,430))
        self.revision = '$Rev: 16 $'[6:-2]
        self.version = '0.1'
        pygame.display.set_caption('PyPebbles %s r%s'%(self.version,self.revision))

        self.background = pygame.Surface(self.screen.get_size())
        self.background = self.background.convert()
        self.background.fill((0,0,0))

        self.screen.blit(self.background, (0,0))
        pygame.display.flip()

        global stones
        stones[1] = load_image('reddot.png',-1)
        stones[2] = load_image('bluedot.png',-1)
        stones[3] = load_image('greendot.png',-1)
        stones[4] = load_image('yellowdot.png',-1)
        stones[5] = load_image('orangedot.png',-1)
        stones[6] = load_image('purpledot.png',-1)
        stones[7] = load_image('lightbluedot.png',-1)
        self.cursor = Cursor()

        #self.font = pygame.font.Font(os.path.join('data','Arial_Black.ttf'), 20)
        self.font = pygame.font.Font(None, 30)
    
        self.clock = pygame.time.Clock()
        self.matrix = [[0 for i in range(8)] for i in range(8)]
        self.delaymatrix = [[4 for i in range(8)] for i in range(8)]
        self.groupmatrix = [[pygame.sprite.GroupSingle() for i in range(8)] for i in range(8)]
        self.allsprites = MyRenderGroup()
        self.anims = MyRenderGroup()
        self.scorelevel = 0
        self.tried_move = 0
        
        self.highscore = HighscoreDisplay(self.screen.get_rect().bottomright,self.font)
        self.anims.add(self.highscore)
        self.anims.add(self.cursor)
        pygame.event.set_blocked(MOUSEMOTION)
    def run(self):
        movement = 1
        self.boardchange = 1
        while 1:
            triples = 0
            for event in pygame.event.get():
                if event.type is QUIT:
                    return
                elif event.type is KEYDOWN and event.key is K_ESCAPE:
                    return
                elif event.type is MOUSEBUTTONDOWN:
                    self.handle_mouse_down()
                elif event.type is MOUSEBUTTONUP:
                    self.handle_mouse_up()
            falling = self.letemfall()
            self.boardchange = falling or self.boardchange
            movement = self.allsprites.update()
            running_anim = self.anims.update()
            if (not movement) and self.boardchange:
                triples = self.look_for_triples()
                if triples:
                    self.tried_move = 0
                    self.boardchange = 1
                else:
                    if self.tried_move:
                        self.undo()
                        self.tried_move = 0
                        self.boardchange = 1
                    else:
                        self.boardchange = 0
            if not (self.boardchange or movement or falling or running_anim):
                event = pygame.event.wait()
                pygame.event.post(event)
            self.allsprites.clear(self.screen,self.background)
            self.anims.clear(self.screen,self.background)
            self.kill_deleted()
            rectlist = self.draw(self.screen)
            pygame.display.update(rectlist)
            self.clock.tick(25)
    def draw(self, screen):
        return self.allsprites.draw(screen)+self.anims.draw(screen)
    def swap(self, pos1, pos2):
        self.swap1 = pos1
        self.swap2 = pos2
        x1, y1 = pos1
        x2, y2 = pos2
        self.matrix[x1][y1], self.matrix[x2][y2] = self.matrix[x2][y2], self.matrix[x1][y1]
        s1 = self.groupmatrix[x1][y1].sprite
        s2 = self.groupmatrix[x2][y2].sprite
        if s1:
            self.groupmatrix[x2][y2].add(s1)
            s1.set_field(x2,y2)
        else:
            self.groupmatrix[x2][y2].empty()
        if s2:
            self.groupmatrix[x1][y1].add(s2)
            s2.set_field(x1,y1)
        else:
            self.groupmatrix[x1][y1].empty()
    def undo(self):
        self.swap(self.swap1,self.swap2)
    def add_stone(self,x,y):
        type = random.randint(1,7)
        self.matrix[x][y] = type
        newstone = PyPebble(type,x,y)
        self.allsprites.add(newstone)
        self.groupmatrix[x][y].add(newstone)
    def check_triple(self,x,y,vertical):
        length = 1
        tx, ty = 0,0
        while 1:
            if vertical:
                tx = x
                ty = y + length
            else:
                tx = x + length
                ty = y
            if tx>7 or ty>7: return length
            if self.matrix[tx][ty] == self.matrix[x][y] != 0:
                length += 1
            else:
                return length
    def look_for_triples(self):
        current_scores = MyRenderGroup()
        triples_found = 0
        for i in range(len(self.matrix)):
            j = 0
            while j < len(self.matrix[i])-2:
                length = self.check_triple(i,j,1)
                if length >= 3:
                    triples_found += 1
                    for k in range(length):
                        self.groupmatrix[i][j+k].sprite.deleted=1
                    score = PyPebblesScore(length,i*50+25,j*50+length*25,self.font)
                    current_scores.add(score)
                j += length
        for i in range(len(self.matrix)):
            j = 0
            while j < len(self.matrix[i])-2:
                length = self.check_triple(j,i,0)
                if length >= 3:
                    triples_found += 1
                    for k in range(length):
                        self.groupmatrix[j+k][i].sprite.deleted=1
                    score = PyPebblesScore(length,j*50+length*25,i*50+25,self.font)
                    current_scores.add(score)
                j += length
        for score in current_scores.spritedict.keys():
            score.multiply(triples_found*(2**self.scorelevel))
            self.highscore.score += score.score
        self.anims.add(current_scores)
        current_scores.empty()
        if triples_found: self.scorelevel += 1
        return triples_found
    def kill_deleted(self):
        for i in range(len(self.matrix)):
            for j in range(len(self.matrix[i])):
                if self.matrix[i][j]:
                    if self.groupmatrix[i][j].sprite.deleted:
                        self.matrix[i][j]=0
                        self.groupmatrix[i][j].sprite.kill()
        for anim in self.anims.spritedict.keys():
            if anim.deleted:
                anim.kill()
    def letemfall(self):
        irange = range(len(self.matrix))
        irange.reverse()
        moving = 0
        for i in irange:
            jrange = range(len(self.matrix[i]))
            jrange.reverse()
            for j in jrange:
                if self.matrix[i][j]==0:
                    moving = 1
                    if self.delaymatrix[i][j]>3:
                        self.delaymatrix[i][j]=0
                        if j==0:
                            self.add_stone(i,j)
                        else:
                            self.swap((i,j),(i,j-1))
                    else:
                        self.delaymatrix[i][j] += 1
        return moving
    def handle_mouse_down(self):
        (x,y) = pygame.mouse.get_pos()
        self.down_field = (x/50,y/50)
        return
    def handle_mouse_up(self):
        (x,y) = pygame.mouse.get_pos()
        up_field = (x/50,y/50)
        if up_field == self.down_field:
            if self.cursor.status == 0:
                self.cursor.set(up_field)
            else:
                if distance(self.cursor.pos,up_field)>1:
                    self.cursor.set(up_field)
                elif distance(self.cursor.pos,up_field)==1:
                    self.swap(self.cursor.pos,up_field)
                    self.boardchange = 1
                    self.scorelevel  = 0
                    self.tried_move  = 1
                    self.cursor.off()
        elif distance(up_field,self.down_field)==1:
            self.swap(self.down_field,up_field)
            self.boardchange = 1
            self.scorelevel  = 0
            self.tried_move  = 1
            self.cursor.off()

class PyPebblesScore(pygame.sprite.Sprite):
    def __init__(self,score,posx,posy,font):
        pygame.sprite.Sprite.__init__(self)
        self.score = score*10
        self.font = font
        self.posx,self.posy = posx,posy
        self.frame = 25
        self.velx,self.vely = 0, -3
        self.deleted = 0
        self.image = self.font.render(str(self.score), 1, (240,240,240))
        self.rect  = self.image.get_rect()
        self.rect.center = (self.posx,self.posy)
    def update(self):
        #if self.frame == 0:
        if self.posy < 0:
            self.deleted = 1
            return 1
        self.posx += self.velx
        self.posy += self.vely
        self.rect.center = (self.posx,self.posy)
        self.frame -= 1
        return 1
    def multiply(self,factor):
        self.score = self.score*factor
        self.image = self.font.render(str(self.score), 1, (240,240,240))
        self.rect  = self.image.get_rect()
        self.rect.center = (self.posx,self.posy)

class HighscoreDisplay(pygame.sprite.Sprite):
    def __init__(self,pos,font):
        pygame.sprite.Sprite.__init__(self)
        self.font = font
        self.pos = pos
        self.score = 0
        self.disp_score = 0
        self.inc = 1
        self.image = self.font.render(str(self.disp_score), 1, (240,240,240))
        self.rect  = self.image.get_rect()
        self.rect.bottomright = self.pos
        self.deleted = 0
    def update(self):
        if self.score > self.disp_score:
            self.inc = (self.score-self.disp_score)/10
            if self.inc < 1: self.inc = 1
            self.disp_score += self.inc
            if self.disp_score > self.score:
                self.disp_score = self.score
            self.image = self.font.render(str(self.disp_score), 1, (240,240,240))
            self.rect  = self.image.get_rect()
            self.rect.bottomright = self.pos
            return 1
        return 0

class Cursor(pygame.sprite.Sprite):
    def __init__(self):
        pygame.sprite.Sprite.__init__(self)
        (self.image, self.rect) = load_image('selection.png',0)
        self.deleted = 0
        self.pos = (-1,-1)
        self.status = 0
        self.drawn = 0
    def set(self,pos):
        self.pos = pos
        self.status = 1
        self.drawn = 0
    def off(self):
        self.pos = (-1,-1)
        self.status = 0
        self.drawn = 0
    def update(self):
        (x,y) = self.pos
        self.rect.topleft = (50*x,50*y)
        if self.drawn:
            return 0
        else:
            self.drawn = 1
            return 1
        
    
        

stones = {}

def main():
    game = PyPebblesGame()


    game.run()


if __name__ == '__main__': main()
#if __name__ == '__main__': profile.run('main()','stats')
#p = pstats.Stats('stats')
#p.strip_dirs()
#p.sort_stats('time')
#p.print_stats()
